from pathlib import Path

import httpx

TRAC_RPC_URL = "https://dev.system/trac/rpc"
TRAC_USERNAME = "trac_login"
TRAC_PASSWORD = "trac_password"


def get_fixture(fixture: Path | str) -> str:
    return (Path(__file__).parent / "fixtures" / fixture).read_text()


RESPONSE_API_VERSION = httpx.Response(
    status_code=httpx.codes.OK,
    text=get_fixture("trac-get-api-version-response.json"),
)



from trac_rpc.models import TracRpcErrorResponse


class TracRpcError(Exception):
    def __init__(self, message: str, *, error: TracRpcErrorResponse = None):
        super().__init__(message)
        self._error = error

    @property
    def error(self) -> TracRpcErrorResponse:
        return self._error



import json
from datetime import datetime

import httpx
import pytest
import respx

from trac_rpc.client import ApiClient
from trac_rpc.models import TracAttachment, TracTicket, TracTicketChangelogEntry, TracTicketProperties

from .utils import get_fixture


@pytest.mark.parametrize(
    ("kwargs", "expected"),
    [
        ({}, "max=0"),
        ({"per_page": 123, "page_number": 456}, "page=456&max=123"),
        ({"query": "status!=closed"}, "status!=closed&max=0"),
        ({"query": "status!=closed", "per_page": 123}, "status!=closed&max=123"),
        ({"query": "status!=closed", "per_page": 123, "page_number": 456}, "status!=closed&page=456&max=123"),
    ],
)
def test_query_tickets(kwargs: dict, expected: str, api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post().mock(
        return_value=httpx.Response(
            status_code=httpx.codes.OK,
            text="""{"error": null, "result": [3, 2, 1], "id": null}""",
        )
    )

    def get_last_request_params() -> str:
        (param,) = json.loads(respx_mock.calls.last.request.content)["params"]
        return param

    assert sorted(api_client.query_tickets(**kwargs)) == [1, 2, 3]
    assert get_last_request_params() == expected


def test_get_ticket_attachments(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post().mock(
        return_value=httpx.Response(
            status_code=httpx.codes.OK,
            text=get_fixture("trac-get-ticket-attachments-response.json"),
        )
    )

    (attachment,) = api_client.get_ticket_attachments(1)
    assert respx_mock.calls.last.request.content == b'{"id":null,"method":"ticket.listAttachments","params":[1]}'

    assert attachment == TracAttachment(
        filename="TracXMLRPC-1.2.0.dev0-py3.13.egg",
        description="Test attachment",
        size=87904,
        timestamp=datetime.fromisoformat("2025-02-27T13:38:04.870563+00:00"),
        author="admin",
    )


def test_get_ticket_changelog(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post().mock(
        return_value=httpx.Response(
            status_code=httpx.codes.OK,
            text=get_fixture("trac-get-ticket-changelog-response.json"),
        )
    )

    changelog = api_client.get_ticket_changelog(1)
    assert respx_mock.calls.last.request.content == b'{"id":null,"method":"ticket.changeLog","params":[1]}'

    assert changelog == [
        TracTicketChangelogEntry(
            timestamp=datetime.fromisoformat("2025-02-27T13:37:11.171873+00:00"),
            author="admin",
            field="comment",
            old_value="1",
            new_value="Test comment",
            permanent=True,
        ),
        TracTicketChangelogEntry(
            timestamp=datetime.fromisoformat("2025-02-27T13:37:11.171873+00:00"),
            author="admin",
            field="owner",
            old_value="somebody",
            new_value="admin",
            permanent=True,
        ),
        TracTicketChangelogEntry(
            timestamp=datetime.fromisoformat("2025-02-27T13:37:11.171873+00:00"),
            author="admin",
            field="status",
            old_value="new",
            new_value="accepted",
            permanent=True,
        ),
        TracTicketChangelogEntry(
            timestamp=datetime.fromisoformat("2025-02-27T13:38:04.870563+00:00"),
            author="admin",
            field="attachment",
            old_value="",
            new_value="TracXMLRPC-1.2.0.dev0-py3.13.egg",
            permanent=False,
        ),
        TracTicketChangelogEntry(
            timestamp=datetime.fromisoformat("2025-02-27T13:38:04.870563+00:00"),
            author="admin",
            field="comment",
            old_value="",
            new_value="Test attachment",
            permanent=False,
        ),
    ]


def test_get_ticket(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post().mock(
        return_value=httpx.Response(
            status_code=httpx.codes.OK,
            text=get_fixture("trac-get-ticket-response.json"),
        )
    )

    assert api_client.get_ticket(1) == TracTicketProperties(
        id=1,
        time_created=datetime.fromisoformat("2025-02-27T13:36:35.856566+00:00"),
        time_changed=datetime.fromisoformat("2025-02-27T13:37:11.171873+00:00"),
        attributes=TracTicket(
            summary="Test summary",
            reporter="admin",
            owner="admin",
            description="Test description\r\n\r\nSecond line",
            type="defect",
            status="accepted",
            priority="major",
            milestone=None,
            component="component1",
            version=None,
            resolution=None,
            keywords=["test1", "test2,test3"],
            cc=[],
            time=datetime.fromisoformat("2025-02-27T13:36:35.856566+00:00"),
            changetime=datetime.fromisoformat("2025-02-27T13:37:11.171873+00:00"),
        ),
    )


def test_get_ticket_last_field_change(api_client: ApiClient, respx_mock: respx.mock):
    respx_mock.post().mock(
        return_value=httpx.Response(
            status_code=httpx.codes.OK,
            text=get_fixture("trac-get-ticket-changelog-response.json"),
        )
    )

    last_comment = api_client.get_ticket_last_field_change(1, "comment")
    assert last_comment == TracTicketChangelogEntry(
        timestamp=datetime.fromisoformat("2025-02-27T13:38:04.870563+00:00"),
        author="admin",
        field="comment",
        old_value="",
        new_value="Test attachment",
        permanent=False,
    )

    last_cc_change = api_client.get_ticket_last_field_change(1, "cc")
    assert last_cc_change is None

    last_owner_change = api_client.get_ticket_last_field_change(1, "owner", "admin")
    assert last_owner_change == TracTicketChangelogEntry(
        timestamp=datetime.fromisoformat("2025-02-27T13:37:11.171873+00:00"),
        author="admin",
        field="owner",
        old_value="somebody",
        new_value="admin",
        permanent=True,
    )



import json

import pytest

from trac_rpc.client import ApiClient, HttpClient
from trac_rpc.models import TracApiVersion, TracComponent, TracMilestone, TracVersion

from .utils import get_fixture

pytestmark = pytest.mark.integration


@pytest.fixture
def api_client():
    return ApiClient(
        rpc_url="http://127.0.0.1:8000/login/rpc",
        http_client=HttpClient(auth=("admin", "admin")),
    )


def test_get_api_version(api_client: ApiClient):
    assert api_client.get_api_version() == TracApiVersion(epoch=1, major=2, minor=0)


def get_list_str_result(fixture_name: str) -> list[str]:
    return json.loads(get_fixture(f"{fixture_name}.json"))["result"]


def test_get_all_components(api_client: ApiClient):
    assert api_client.get_all_components() == get_list_str_result("trac-get-all-components-response")


def test_get_all_milestones(api_client: ApiClient):
    assert api_client.get_all_milestones() == get_list_str_result("trac-get-all-milestones-response")


def test_get_all_priorities(api_client: ApiClient):
    assert api_client.get_all_priorities() == get_list_str_result("trac-get-all-priorities-response")


def test_get_all_resolutions(api_client: ApiClient):
    assert api_client.get_all_resolutions() == get_list_str_result("trac-get-all-resolutions-response")


def test_get_all_severities(api_client: ApiClient):
    assert api_client.get_all_severities() == get_list_str_result("trac-get-all-severities-response")


def test_get_all_statuses(api_client: ApiClient):
    assert api_client.get_all_statuses() == get_list_str_result("trac-get-all-statuses-response")


def test_get_all_types(api_client: ApiClient):
    assert api_client.get_all_types() == get_list_str_result("trac-get-all-types-response")


def test_get_all_versions(api_client: ApiClient):
    assert api_client.get_all_versions() == get_list_str_result("trac-get-all-versions-response")


def test_get_milestone(api_client: ApiClient):
    assert api_client.get_milestone("milestone2") == TracMilestone(
        name="milestone2",
        description="",
        completed=None,
        due=None,
    )


def test_get_component(api_client: ApiClient):
    assert api_client.get_component("component2") == TracComponent(
        name="component2",
        owner="somebody",
        description="",
    )


def test_get_version(api_client: ApiClient):
    assert api_client.get_version("2.0") == TracVersion(
        name="2.0",
        released=None,
        description="",
    )


def test_get_all_wiki_pages(api_client: ApiClient):
    assert sorted(api_client.get_all_wiki_pages()) == sorted(get_list_str_result("trac-get-all-wiki-pages-response"))


def test_wiki_to_html(api_client: ApiClient):
    assert (
        api_client.wiki_to_html("'''bold''', ''italic'', `monospaced`")
        == """\
<p>
<strong>bold</strong>, <em>italic</em>, <code>monospaced</code>
</p>
"""
    )



from typing import Any

import pytest

from trac_rpc.validators import (
    validate_comma_separated,
    validate_in_set,
    validate_space_or_comma_separated,
    validate_space_separated,
)


@pytest.mark.parametrize(
    ("value", "allowed", "optional", "raises"),
    [
        (1, {1, 2, 3}, False, False),
        (5, {1, 2, 3}, False, True),
        (None, {1, 2, 3}, False, True),
        (None, {1, 2, 3}, True, False),
    ],
)
def test_validate_in_set(value: int | None, allowed: set[int], optional: bool, raises: bool):
    if raises:
        with pytest.raises(ValueError, match=".+ is not in .+"):
            validate_in_set(value, allowed, optional)
    else:
        assert validate_in_set(value, allowed, optional) == value


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (123, 123),
        ("", []),
        (" ", []),
        ("  ", []),
        ("  test", ["test"]),
        ("  test ", ["test"]),
        ("  test1   test2 ", ["test1", "test2"]),
        ("  test1   test2,test3 ", ["test1", "test2,test3"]),
        ("  test1   test2, test3 ", ["test1", "test2,", "test3"]),
    ],
)
def test_validate_space_separated(value: Any, expected: Any):
    assert validate_space_separated(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (123, 123),
        ("", []),
        (" ", []),
        ("  ", []),
        ("  test", ["test"]),
        ("  test ", ["test"]),
        ("  test1   test2 ", ["test1   test2"]),
        ("  test1   test2,test3 ", ["test1   test2", "test3"]),
        ("  test1   test2, test3 ", ["test1   test2", "test3"]),
    ],
)
def test_validate_comma_separated(value: Any, expected: Any):
    assert validate_comma_separated(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (123, 123),
        ("", []),
        (" ", []),
        ("  ", []),
        ("  test", ["test"]),
        ("  test ", ["test"]),
        ("  test1   test2 ", ["test1", "test2"]),
        ("  test1   test2,test3 ", ["test1   test2", "test3"]),
        ("  test1   test2, test3 ", ["test1   test2", "test3"]),
    ],
)
def test_validate_space_or_comma_separated(value: Any, expected: Any):
    assert validate_space_or_comma_separated(value) == expected
